#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include "defimport.h"

#define isblank isspace
#define alloca __chkstk

char **_argv;
int _argc;
static char** CmdFiles = NULL;
static char** CmdOutput = NULL;
static char** CmdFlags = NULL;
static char* Output = NULL;

IMPORTS *Imports = NULL;

char **GetSwitchData(char *Switch) {
	char Parameter[256] = {0};
	char Temp[256] = {0};
	int IsSwitch = 0;

	for (int i = 1; i < _argc; i++) {
		if (_argv[i][0] == '-') {
			IsSwitch = 1;
		} else {
			if (IsSwitch) {
				IsSwitch = 0;
			} else {
				/* handle non '-' Switch case */
				if (Switch[0] != '-') {
					strcpy(Parameter, _argv[i]);
					break;
				}
			}
		}

		if (strcmp(_argv[i], Switch) == 0) {
			if (i + 1 < _argc)
				strcpy(Parameter, _argv[i + 1]);
			else
				*Parameter = '\0';
			break;
		}
	}

	strcpy(Temp, Parameter);

	int TokenCount = 0;
	char *Token = strtok(Temp, ",+;");

	while (Token != NULL) {
		TokenCount++;
		Token = strtok(NULL, ",+;");
	}

	Token = strtok(Parameter, ",+;");

	if (TokenCount) {
		/* Allocate Array */

		char **data = malloc((1 + TokenCount) * sizeof(char*));

		/* Set parameter count */

		*(int*) data = TokenCount;

		for (int i = 1; i <= TokenCount; i++) {
			data[i] = malloc(strlen(Token) + sizeof(char));
			strcpy(data[i], Token);
			Token = strtok(NULL, ",+;");
		}
		return data;
	}
	return NULL;
}

int CreationFlagsSet(char* Data[], char* Flag) {
	if (Data == NULL) return 0;
	if (Data[0] > 0) {
		for (int i = 1; i <= (int) Data[0]; i++) {
			if (strcmp (Data[i], Flag) == 0) return 1;
		}
	}
	return 0;
}

char * strtrim (char *s) {
	if (s == NULL) return 0;
	while (*s && (isblank (*s) || iscntrl (*s))) {
		s++;
	}
	int l = strlen (s) - 1;
	while (l > 0 && (isblank (s[l]) || iscntrl (s[l]))) {
		s[l] = '\0';
		l--;
	}
	return s;
}

/* USE THIS IF NOT ALREADY EXISTS IN STRING.H (NOT DEFINED IN ISO C99)
char *strdup(char *s) {
	char *n = malloc(strlen (s));
	strcpy(n, s);
	return n;
}*/

IMPORT_ENTRY *AddLibraryImport (char* LibName) {
	IMPORTS *pimp,*imp;
	IMPORT_ENTRY *p,*n = malloc (sizeof (IMPORT_ENTRY));
	if (Imports == NULL) {
		Imports = malloc (sizeof (IMPORTS));
		Imports->LibName = strdup (LibName);
		Imports->Items = NULL;
		Imports->Next = NULL;
	}
	imp = Imports;
	while (imp) {
		if (strcmp (imp->LibName, LibName) == 0) {
			p = imp->Items;
			imp->Items = n;
			n->Next = p;
			return n;
		}
		imp = imp->Next;
	}
	pimp = Imports->Next;
	imp = malloc (sizeof (IMPORTS));
	imp->LibName = strdup (LibName);
	imp->Next = pimp;
	imp->Items = n;
	Imports->Next = imp;
	n->Next = NULL;
	return n;
}

int EnumImportLibs (int (*EnumProc)(IMPORTS* Lib, void* Data), void* Data) {
	IMPORTS *imp = Imports;
	int idx = 0;
	while (imp) {
		idx++;
		if (EnumProc (imp, Data) == 0) return idx;
		imp = imp->Next;
	}
	return idx;
}

int EnumImports (int (*EnumProc)(IMPORTS *Lib, IMPORT_ENTRY *n, void* Data), IMPORTS *Lib, void* Data) {
	IMPORT_ENTRY *i = Lib->Items;
	int idx = 0;
	while (i) {
		idx++;
		if (EnumProc (Lib, i, Data) == 0) return idx;
		i = i->Next;
	}
	return idx;
}

void ClearLibraryList () {
	IMPORTS *nimp, *imp = Imports;
	IMPORT_ENTRY *i,*n;
	while (imp) {
		nimp = imp->Next;
		free (imp->LibName);
		i = imp->Items;
		while (i) {
			n = i->Next;
			if (i->InternalName) free (i->InternalName);
			if (i->ExternalName) free (i->ExternalName);
			free (i);
			i = n;
		}
		free (imp);
		imp = nimp;
	};
}

char *MangleName (char *s) {
	char *n = s;
	while (s && *s) {
		if (!isalnum (*s)) *s = '_';
		s++;
	}
	return n;
}

int IsFncNameValid (char *s) {
	if (s == NULL) return 0;
	/* Don't really know which characters allowed to be used as symbol name in object files (thinking on c++ symbol names) */
	if (isdigit (*s) || *s < 32 /*|| *s > 127 */|| *s == '.') {
		return 0;
	} else {
		while (*s) {
			if (*s < 32 || /* *s > 127 || */*s == '.') return 0;
			s++;
		}
	}
	return 1;
}

int ParseDefFile (char* File) {
	char LineBuffer[180], CurLibrary[256], Prefix[16];
	char *Line;
	int LineNo = 0;
	/* Set default current library which will be used if LIBRARY keyword is missed */
	strcpy(CurLibrary, File);
	strcat(CurLibrary, ".dll");

	FILE *f = fopen (File, "r");
	if (f != NULL) {
		while (fgets (LineBuffer, sizeof (LineBuffer), f)) {
			LineNo++;
			Line = LineBuffer;
			Line = strtrim(LineBuffer); // trim all spaces

			if (*Line == ';') continue; // Ignore comments
			if (*Line == '\0') continue; // Ignore empty lines

			char *Name = Line;
			char *Value = strchr (Line, '=');
			if (Value) {
				*Value = '\0';
				Value++;
				Value = strtrim (Value);
			} else {
				Value = "";
			}
			Name = strtrim (Name);

			/* Test keywords */
			if (strcmp (Name, ".OUTPUT") == 0) {
				if (Output) free (Output);
				Output = malloc (strlen (Value) + 2 + 1);
				strcpy (Output, Value);
				char *s = strrchr (Output, '.');
				if (s) {
					strcpy (s, ".s");
				} else {
					strcat (s, ".s");
				}
			} else if (strcmp(Name, ".PREFIX") == 0) {
				strncpy (Prefix, Value, sizeof(Prefix));
			} else if (strcmp(Name, ".LIBRARY") == 0) {
				strcpy (CurLibrary, Value);
			} else {
				if (strlen (Value) == 0) {
					printf ("Error in line %d: No external library name was specified for '%s'\n",LineNo,Name);
				} else if (strlen (Name) == 0) {
					printf ("Error in line %d: No internal symbol name was specified for '%s'\n",LineNo,Value);
				} else {
					if (IsFncNameValid (Name)) {
						if (IsFncNameValid (Value)) {
							/* import by string name */
							IMPORT_ENTRY *n = AddLibraryImport (CurLibrary);
							n->InternalName = malloc (strlen (Name) + strlen (Prefix) + 1);
							strcpy (n->InternalName, Prefix);
							strcat (n->InternalName, Name);
							n->ExternalName = strdup (Value);
							n->PrefixWidth = strlen (Prefix);
							n->Ordinal = 0;
						} else {
							if (isdigit (*Value)) {
								/* Import by ordinal */
								IMPORT_ENTRY *n = AddLibraryImport (CurLibrary);
								n->InternalName = malloc (strlen (Name) + strlen (Prefix) + 1);
								strcpy (n->InternalName, Prefix);
								strcat (n->InternalName, Name);
								n->ExternalName = NULL;
								n->PrefixWidth = strlen (Prefix);
								n->Ordinal = strtol (Value, NULL, 0);
							} else {
								printf("Error in line %d: External library name not in valid format '%s'\n", LineNo, Value);
							}
						}
					} else {
						printf("Error in line %d: Symbol name not in valid format '%s'\n", LineNo, Name);
					}
				}
			}
		}
		fclose (f);
	} else {
		printf ("Error: '%s' %s!\n",File,strerror (errno));
	}
	return 0;
}

void finalize () {
	if (CmdFiles) free (CmdFiles);
	if (CmdOutput) free (CmdOutput);
	if (CmdFlags) free (CmdFlags);
	if (Output) free (Output);
	if (Imports) ClearLibraryList();
}

void DoPrintHelp() {
	printf ("DEF file importer (can not read standard DEF files) Version 1.1\n");
	printf ("Produced initially for Ozon Operating System (OOS)\n");
	printf ("Usage: defimport.exe [options] files\n");
	printf (" -o FILE        Output assembler file\n");
	printf (" -f FLAGS       Creation flags:\n");
	printf ("    jmp         Also Create indirect DLL calls (JMP DS:[__imp__fnc])\n");
	printf ("    as          Assemble output file (create object file)\n");
	printf ("    d           Delete immediate files (.s files), only valid with 'as' flag\n");
	printf ("\nUsage Example: defimport -f jmp,as,d opengl32.def\n");
	printf ("\nLicence: None (free for all)\n");
	printf ("Author: Benjamin Kalytta\n");
}

int main (int argc, char* argv[]) {
	_argc = argc;
	_argv = argv;
	if (argc == 1) {
		DoPrintHelp();
		return -1;
	}

	atexit (finalize);

	CmdFlags = GetSwitchData ("-f"); // creation flags
	CmdFiles = GetSwitchData (""); // input def files
	CmdOutput = GetSwitchData ("-o"); // output file (can be overriden by .OUTPUT keyword in def file)

	if (CmdFiles == NULL) {
		printf("Error: Please specify at least one DEF file!\n");
		return -1;
	}

	/* Create output file name */
	if (CmdOutput) {
		Output = malloc (strlen (CmdOutput[1]) + 2 + 1);
		strcpy (Output, CmdOutput[1]);
	} else {
		Output = malloc (strlen (CmdFiles[1]) + 2 + 1);
		strcpy(Output, CmdFiles[1]);
	}

	for (int i = 1; i <= (int) CmdFiles[0]; i++) {
		ParseDefFile (CmdFiles[i]);
	}

	/* add file extension */
	char *s = strrchr(Output, '.');
	if (s) {
		strcpy (s, ".s");
	} else {
		strcat (s, ".s");
	}

	FILE *fo = fopen(Output, "w");
	if (fo == NULL) printf("defimport: Error: '%s': %s\n", Output, strerror(errno));

	int SymbolEnumProc(IMPORTS *Lib, IMPORT_ENTRY *n, void* Data) {
		char MgName[256] = {0};
		char Prefix[16];
		strncpy (MgName, Lib->LibName, sizeof (MgName));
		MangleName (MgName);
		strncpy(Prefix, n->InternalName, MIN (sizeof (Prefix), n->PrefixWidth));
		Prefix[n->PrefixWidth] = '\0';

		char Buffer[4096];

		void CreateIndirectJumps() {
			snprintf(Buffer, sizeof(Buffer),
				"\n"
				".global %s%s\n"
				"%s%s:\n"
				"	jmp [%s_imp__%s]\n",
				Prefix, &n->InternalName[n->PrefixWidth],
				Prefix, &n->InternalName[n->PrefixWidth],
				Prefix, &n->InternalName[n->PrefixWidth]
			);
			fputs(Buffer, fo);
		}

		void CreateGlobals() {
			snprintf(Buffer, sizeof(Buffer),
				"\n"
				".global %s_imp__%s\n",
				Prefix, &n->InternalName[n->PrefixWidth]
			);
			fputs(Buffer, fo);
		}

		void CreateIData4() {
			if (n->ExternalName) {
				snprintf(Buffer, sizeof(Buffer),
					"	.rva %s_sym_%s_%s\n",
					Prefix, MgName, &n->InternalName[n->PrefixWidth]
				);
			} else {
				snprintf(Buffer, sizeof(Buffer),
					"	.long %#.8x\n",
					n->Ordinal + 0x80000000
				);
			}
			fputs(Buffer, fo);
		}

		void CreateIData5() {
			snprintf(Buffer, sizeof(Buffer),
				"%s_imp__%s:\n"
				"	.rva %s_sym_%s_%s\n",
				Prefix, &n->InternalName[n->PrefixWidth],
				Prefix, MgName, &n->InternalName[n->PrefixWidth]
			);
			fputs(Buffer, fo);
		}

		void CreateIData6() {
			if (n->ExternalName) {
				snprintf(Buffer, sizeof(Buffer),
					"%s_sym_%s_%s:\n"
					"	.short 0\n"
					"	.asciz \"%s\"\n",
					Prefix, MgName, &n->InternalName[n->PrefixWidth],
					n->ExternalName
				);
			} else {
				snprintf(Buffer, sizeof(Buffer),
					"%s_sym_%s_%s:\n"
					"	.short 0\n"
					"	.asciz \"\"\n",
					Prefix, MgName, &n->InternalName[n->PrefixWidth]
				);
			}
			fputs(Buffer, fo);
		}

		switch ((int) Data) {
			case 0:
				CreateIndirectJumps();
				break;
			case 1:
				CreateGlobals();
				break;
			case 2:
				CreateIData4();
				break;
			case 3:
				CreateIData5();
				break;
			case 4:
				CreateIData6();
				break;
		}
		return 1;
	}

	int SymbolEnumLibProc(IMPORTS *Lib, void* Data) {
		char Buffer[4096];
		char MgName[256] = {0};
		strncpy (MgName, Lib->LibName, sizeof (MgName));
		MangleName (MgName);

		if (CreationFlagsSet (CmdFlags, "jmp")) {
			EnumImports(SymbolEnumProc, Lib, (void*) 0); // Create indirect jumps
		}

		EnumImports(SymbolEnumProc, Lib, (void*) 1); // Create global symbols

		/* Create import directory (.idata$2 section) */
		snprintf(Buffer, sizeof(Buffer),
			"\n"
			".section .idata$2\n"
			"	.rva ilt_%s\n"
			"	.long 0\n"
			"	.long 0\n"
			"	.rva %s\n"
			"	.rva iat_%s\n"
			"\n"
			".section .idata$3\n"
			"	.long 0\n"
			"	.long 0\n"
			"	.long 0\n"
			"	.long 0\n"
			"	.long 0\n"
			"\n"
			".section .idata$4\n"
			"ilt_%s:\n",
			MgName, MgName, MgName, MgName
		);
		fputs(Buffer, fo);

		EnumImports(SymbolEnumProc, Lib, (void*)2); // Create .idata$4 section
		fputs("	.long 0\n", fo); // indication that .idata$4 for current dll ends here

		snprintf (Buffer, sizeof(Buffer),
			"\n"
			".section .idata$5\n"
			"iat_%s:\n",
			MgName
		);
		fputs(Buffer, fo);
		EnumImports(SymbolEnumProc, Lib, (void*)3); // Create .idata$5 section
		fputs("	.long 0\n", fo); // indication that .idata$5 for current dll ends here

		snprintf (Buffer, sizeof(Buffer),
			"\n"
			".section .idata$6\n"
		);
		fputs(Buffer, fo);
		EnumImports(SymbolEnumProc, Lib, (void*)4); // Create .idata$6 section

		snprintf (Buffer, sizeof(Buffer), // Create .idata$7 section (import dll name)
			"\n"
			".section .idata$7\n"
			"%s:\n"
			"	.asciz \"%s\"\n",
			MgName, Lib->LibName
		);
		fputs(Buffer, fo);
		return 1;
	}

	if (fo == NULL) fo = stdout;

	fprintf(fo, ".intel_syntax noprefix\n\n");
//	fprintf(fo, ".syntax intel\n\n");

	EnumImportLibs(SymbolEnumLibProc, 0);

	fclose(fo);

	if (CreationFlagsSet (CmdFlags, "as")) {
		/* call gnu assembler */
		char SysCmd[256];
		char a[256], b[256];
		char *s = strrchr(Output, '.');
		if (s) {
			*s = '\0';
		}
		snprintf(a, 64, "%s.o", Output);
		snprintf(b, 64, "%s.s", Output);
		snprintf(SysCmd, 256 , "as -o %s %s", a, b);
		system(SysCmd);

		if (CreationFlagsSet (CmdFlags, "d")) {
			snprintf(SysCmd, 256 , "del %s", b);
			system(SysCmd);
		}

	}
	return 0;
}
